/*============================================================================
  File:     03_Statistics and Invalidation.sql

  SQL Server Versions: 2012 onwards
------------------------------------------------------------------------------
  Written by Erin Stellato, SQLskills.com
  
  (c) 2015, SQLskills.com. All rights reserved.

  For more scripts and sample code, check out 
    http://www.SQLskills.com

  You may alter this code for your own *non-commercial* purposes. You may
  republish altered code as long as you include this copyright and give due
  credit, but you must obtain prior permission before blogging this code.
  
  THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
  ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED 
  TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
  PARTICULAR PURPOSE.
============================================================================*/

USE [AdventureWorks2014];
GO

/*
	"all of the sudden" there's a perf issue with a SP, what might you do?
	*enable execution plan
*/
SET STATISTICS IO ON;
GO

EXECUTE Sales.usp_GetProductInfo 739;
GO

EXECUTE Sales.usp_GetProductInfo 840; 
GO

EXECUTE Sales.usp_GetProductInfo 890;
GO

EXECUTE Sales.usp_GetProductInfo 871;
GO

EXECUTE Sales.usp_GetProductInfo 900;
GO

EXECUTE Sales.usp_GetProductInfo 870;
GO

/*
	Are stats out of date?
*/
SELECT
    sch.name + '.' + so.name AS "Table",
    ss.name AS "Statistic",
	ss.auto_Created AS "Auto Created?",
	ss.user_created AS "User Created?",
	ss.has_filter AS "Filtered?", 
    ss.filter_definition AS "Filter Definition", 
    sp.last_updated AS "Stats Last Updated", 
    sp.rows AS "Rows in Table", 
    sp.rows_sampled AS "Rows Sampled", 
    sp.unfiltered_rows AS "Unfiltered Rows",
	sp.modification_counter AS "Row Modifications",
	sp.steps AS "Histogram Steps"
FROM sys.stats ss
JOIN sys.objects so ON ss.object_id = so.object_id
JOIN sys.schemas sch ON so.schema_id = sch.schema_id
CROSS APPLY sys.dm_db_stats_properties(so.object_id, ss.stats_id) AS sp
WHERE so.object_id = object_id('Sales.SalesOrderDetail')
ORDER BY ss.user_created, ss.auto_created, ss.has_filter;
GO

/*
	Not a full scan, maybe that's our issue, let's update stats!
*/
UPDATE STATISTICS Sales.SalesOrderDetail WITH FULLSCAN;


/*
	Re-run the SP...did we fix perf?
*/
EXECUTE Sales.usp_GetProductInfo 870;
GO

EXECUTE Sales.usp_GetProductInfo 739;
GO

EXECUTE Sales.usp_GetProductInfo 840; 
GO

EXECUTE Sales.usp_GetProductInfo 890;
GO

EXECUTE Sales.usp_GetProductInfo 871;
GO

EXECUTE Sales.usp_GetProductInfo 900;
GO




/*
	uhhhhh...no.  Rats.  Let's just clear procedure cache...
	
	:'(

	Very sad panda
*/
DBCC FREEPROCCACHE;
GO

/*
	re-run the SP
*/
EXECUTE Sales.usp_GetProductInfo 870;
GO

EXECUTE Sales.usp_GetProductInfo 739;
GO

EXECUTE Sales.usp_GetProductInfo 840; 
GO

EXECUTE Sales.usp_GetProductInfo 890;
GO

EXECUTE Sales.usp_GetProductInfo 871;
GO

EXECUTE Sales.usp_GetProductInfo 900;
GO


/*
	Still bad!!!  (and we probably spiked CPU on the server!)

	STOP.  Why?  What's going on here?

	What are some things you could do here to troubleshoot?

	Free procedure cache? (tried that, it didn't work)
	Remove that one plan from procedure cache (maybe?)
	Restart the instance?  (uh, no...that's a sledgehammer, that's not troubleshooting :)
	Look a little closer at the plan?  YES.  Let's understand WHY this query is a problem before we DO anything.

*/





/*
	Find the plan for the query in question in the cache
	Let's make sure that the plan we see in SSMS is the same
	as what users are seeing!
*/
SELECT qs.execution_count, s.text, qs.query_hash, qs.query_plan_hash, cp.size_in_bytes/1024 [PlanSizeKB], qp.query_plan, qs.plan_handle
FROM sys.dm_exec_query_stats qs
CROSS APPLY sys.dm_exec_query_plan (qs.plan_handle) qp
CROSS APPLY sys.dm_exec_sql_text(qs.plan_handle) s
JOIN sys.dm_exec_cached_plans cp ON qs.plan_handle = cp.plan_handle
WHERE s.text LIKE '%SalesOrderDetail%';

/*
	What operations are occurring?
	What values were used for compilation?
		look in the plan for the compiled value!
	What does the distribution of data look like for ProductID?
		compare value from histogram with table dist
*/
DBCC SHOW_STATISTICS ('Sales.SalesOrderDetail', IX_SalesOrderDetail_ProductID);
GO

SELECT ProductID, COUNT(*)
FROM Sales.SalesOrderDetail
GROUP BY ProductID
ORDER BY COUNT(*) DESC;

/*
	Run some individual queries with literal values...as we did in the beginning
	*enable execution plan
*/
SET STATISTICS IO ON;

SELECT [ProductID], [OrderQty]
FROM [Sales].[SalesOrderDetail]
WHERE [ProductID] = 732
ORDER BY [ProductID];

SELECT [ProductID], [OrderQty]
FROM [Sales].[SalesOrderDetail]
WHERE [ProductID] = 870
ORDER BY [ProductID];

SELECT [ProductID], [OrderQty]
FROM [Sales].[SalesOrderDetail]
WHERE [ProductID] = 739
ORDER BY [ProductID];

SELECT [ProductID], [OrderQty]
FROM [Sales].[SalesOrderDetail]
WHERE [ProductID] = 840
ORDER BY [ProductID];






/*
	Ok, I've got a parameter sensitivity issue, so now I can understand WHY the perf issue is intermittent
	But why doesn't updating stats address it?
	Step through again, looking at plan cache along the way
*/


/*
	what exists in cache currently?
*/
SELECT qs.execution_count, s.text, qs.query_hash, qs.query_plan_hash, cp.size_in_bytes/1024 [PlanSizeKB], qp.query_plan, qs.plan_handle
FROM sys.dm_exec_query_stats qs
CROSS APPLY sys.dm_exec_query_plan (qs.plan_handle) qp
CROSS APPLY sys.dm_exec_sql_text(qs.plan_handle) s
JOIN sys.dm_exec_cached_plans cp ON qs.plan_handle = cp.plan_handle
WHERE s.text LIKE '%SalesOrderDetail%';

/*
	Update stats, this should invalidate our plan so we get a new one
*/
UPDATE STATISTICS Sales.SalesOrderDetail WITH FULLSCAN;


/*
	Verify they updated
*/
DBCC SHOW_STATISTICS ('Sales.SalesOrderDetail', IX_SalesOrderDetail_ProductID);
GO

/*
	Re-run the SP...
*/
EXECUTE Sales.usp_GetProductInfo 739;
GO

EXECUTE Sales.usp_GetProductInfo 840; 
GO

EXECUTE Sales.usp_GetProductInfo 890;
GO

EXECUTE Sales.usp_GetProductInfo 871;
GO

EXECUTE Sales.usp_GetProductInfo 900;
GO

EXECUTE Sales.usp_GetProductInfo 870;
GO

/*
	What's the execution count?
*/
SELECT qs.execution_count, s.text, qs.query_hash, qs.query_plan_hash, cp.size_in_bytes/1024 [PlanSizeKB], qp.query_plan, qs.plan_handle
FROM sys.dm_exec_query_stats qs
CROSS APPLY sys.dm_exec_query_plan (qs.plan_handle) qp
CROSS APPLY sys.dm_exec_sql_text(qs.plan_handle) s
JOIN sys.dm_exec_cached_plans cp ON qs.plan_handle = cp.plan_handle
WHERE s.text LIKE '%SalesOrderDetail%';


/*
	So why isn't an update to statistics invalidating my plan????

	It's because of a change in 2012...

	Let's update one row of data, then update stats
*/
UPDATE [Sales].[SalesOrderDetail]
SET [ProductID] = 915
WHERE [ProductID] = 911
AND [SalesOrderID] = 53488
AND [SalesOrderDetailID] = 47780;

/*
	Update stats again, this time it should invalidate our plan so we get a new one
*/
UPDATE STATISTICS Sales.SalesOrderDetail WITH FULLSCAN;


/*
	Verify they updated
*/
DBCC SHOW_STATISTICS ('Sales.SalesOrderDetail', IX_SalesOrderDetail_ProductID);
GO

/*
	Re-run the SP
*/
EXECUTE Sales.usp_GetProductInfo 915;
GO

/*
	Verify the execution count
*/
SELECT qs.execution_count, s.text, qs.query_hash, qs.query_plan_hash, cp.size_in_bytes/1024 [PlanSizeKB], qp.query_plan, qs.plan_handle
FROM sys.dm_exec_query_stats qs
CROSS APPLY sys.dm_exec_query_plan (qs.plan_handle) qp
CROSS APPLY sys.dm_exec_sql_text(qs.plan_handle) s
JOIN sys.dm_exec_cached_plans cp ON qs.plan_handle = cp.plan_handle
WHERE s.text LIKE '%SalesOrderDetail%';



/*
	What about 2008R2 and below?

*/
